//#define DEBUG

#include "myomnipotentmanagerframework.h"
#include <vector>
#include <stack>
#include <cstdlib>
#include <numeric>
#include <random>
#include "scoreinfo.h"

using namespace std;
typedef pair<int, int> edge;

constexpr int VISIT = 473895279, ANSWER = 1979520194;

#ifdef DEBUG
  #define VISITING_ALICE "Alice must not call visit"
  #define TOO_MANY_VISITS "Bob called visit too often"
  #define WRONG_ANSWER "Wrong answer"
  #define INVALID_ALICE "Return value of alice invalid"
  #define TOO_MANY_TIES "Too many ties in one room"
  #define INVALID_VISIT "Invalid call to visit"
  #define CORRECT(x,y) "Correct: %d extra visits, Kmax = %d", (x), (y)
  #define SECURITY_VIOLATION "Security violation!"
#else
  #define VISITING_ALICE "Not correct"
  #define TOO_MANY_VISITS "Not correct"
  #define WRONG_ANSWER "Not correct"
  #define INVALID_ALICE "Not correct"
  #define TOO_MANY_TIES "Not correct"
  #define INVALID_VISIT "Not correct"
  #define CORRECT(x,y) y <= 1? "Correct" : "Partially correct"
  #define SECURITY_VIOLATION "Security violation!"
#endif

#define write(x) _write_helper(__LINE__, x)
template<typename T> void _write_helper(int line, T t) {
  fwrite(&t, sizeof(t), 1, fcommout[0]);
  fflush(fcommout[0]);

  #ifdef DEBUG_
    cerr << "[line" << line << "] " << t << flush;
  #endif
}

template<typename T> T read_or_fail() {
  T x;
  if (fread(&x, sizeof(x), 1, fcommin[0]) != 1)
    result(0.0, SECURITY_VIOLATION);
  return x;
}

mt19937_64 rng;

class abstract_base_adversary {
  public:
    virtual ~abstract_base_adversary() {}

    virtual pair<vector<edge>, int> for_alice() = 0;
    virtual tuple<vector<edge>, int, int, int> for_bob() = 0;

    virtual void answer_alice(vector<int> v) = 0;
    virtual bool answer_bob() = 0;
    virtual int visit(int) = 0;
    virtual int get_N() = 0;
};

class basic_adversary : public abstract_base_adversary {
  public:
    basic_adversary(vector<edge> E, int s, int d) :
     edges(E), N(E.size() + 1), dest(d), curr_node(s), T(N) {
      for(auto [a, b] : edges)
        T[a].push_back(b), T[b].push_back(a);
    }

    basic_adversary(vector<edge> E) :
     basic_adversary(E, uniform_int_distribution<>(0, E.size())(rng),
                        uniform_int_distribution<>(0, E.size())(rng)) {}

    virtual ~basic_adversary() {}

    virtual pair<vector<edge>, int> for_alice() { return {edges, dest}; }
    virtual tuple<vector<edge>, int, int, int> for_bob() {
      return {edges, curr_node, ids[curr_node], dist(curr_node, dest)};
    }

    virtual void answer_alice(vector<int> v) { ids = v; }
    virtual bool answer_bob() { return dest == curr_node; } // should we check whether the solution is unique?
    virtual int visit(int v) {
      if(not connected(curr_node, v)) return -1;
      curr_node = v;
      return ids[curr_node];
    }
    virtual int get_N() { return N; }

  protected:
    vector<int> ids;
    vector<edge> edges;
    int N;
    int dest;
    int curr_node;
    vector<vector<int>> T;

    virtual int dist(int a, int b) {
      stack<int> s; s.push(a);
      vector<int> dist(T.size(), -1); dist[a] = 0;

      while(not s.empty()) {
        int x = s.top(); s.pop();

        for(int w : T[x])
          if(dist[w] < 0)
            dist[w] = dist[x] + 1, s.push(w);
      }

      return dist[b];
    }

    virtual bool connected(int v, int w) {
      for(int x : T[v])
        if(x == w) return true;
      return false;
    }

    friend class libra_adversary;
    virtual int get_id() { return ids[curr_node]; }
};

class symmetric_base_adversary : public basic_adversary {
  public:
    symmetric_base_adversary(vector<edge> E, int start) : basic_adversary(E) {
      curr_node = start;
      pi.resize(N, -1);
      pi[start] = 0;
    }
    virtual ~symmetric_base_adversary() {}

    virtual bool answer_bob() {
      return pi[curr_node] == dest;
    }

    virtual int visit(int) = 0;

  protected:
    vector<int> pi;

    static int parent(int x) { return (x - 1) / 2; }
    static int left(int x)   { return 2 * x + 1; }
    static int right(int x)  { return 2 * x + 2; }

    static bool in_subtree(int root, int x) {
      while(x > root) x = parent(x);
      return x == root;
    }
};

class symmetric_adversary : public symmetric_base_adversary {
  public:
    symmetric_adversary(vector<edge> E, bool up, bool off) : symmetric_base_adversary(E, 0) {
      assert(__builtin_popcount(N + 1) == 1);
      for(int i = 1; i < N; ++i)
        assert(edges[i - 1].second == parent(i) and edges[i - 1].first == i);

      dest = N - 1; if(up) dest = parent(dest);
      if(off) {
        vector<int> l; l.push_back(1);
        vector<char> isl(N, false); isl[1] = true;

        for(int i = 2; i < N; ++i) {
          int j = parent(i);

          if(isl[j]) {
            l.push_back(i);
            isl[i] = true;
            pi[i] = i;
          }
        }

        curr_node = l[uniform_int_distribution<>(0, l.size() - 1)(rng)];
        pi[1] = 1; pi[2] = 2;
        assert(pi[0] == 0);
      }
    }

    virtual int visit(int v) {
      assert(pi[curr_node] != -1);

      if(v == parent(curr_node)) {
        curr_node = v;
        assert(pi[curr_node] != -1);
        return ids[pi[curr_node]];
      }

      if(v != left(curr_node) and v != right(curr_node)) return -1;

      if(pi[v] == -1) {
        int l = left(curr_node), r = right(curr_node);
        assert(pi[l] == -1 and pi[r] == -1);
        assert(pi[curr_node] != -1);

        pi[l] = left(pi[curr_node]);
        pi[r] = right(pi[curr_node]);

        if(rng() % 2)               swap(pi[l], pi[r]);
        if(in_subtree(pi[v], dest)) swap(pi[l], pi[r]);
      }

      curr_node = v;
      return ids[pi[curr_node]];
    }

  protected:
    friend class libra_adversary;
    virtual int get_id() { return ids[curr_node]; }
};

enum src_strategy { STD, RANDOM, TAIL };

class symmetric_app_adversary : public symmetric_base_adversary {
  public:
    symmetric_app_adversary(vector<edge> E, bool up, src_strategy spos) : symmetric_base_adversary(E, E.size()), M(1) {
      while(M < N) M <<= 1; M >>= 1; --M;

      for(int i = 1; i < M; ++i)
        assert(edges[i - 1] == make_pair(i, parent(i)));
      for(int i = M; i < N - 1; ++i)
        assert(edges[i].first >= M and edges[i].second >= M);
      assert(edges[M - 1] == make_pair(0, N - 1));

      dest = M - 1; if(up) dest = parent(dest);
      for(int i = M; i < N; ++i) pi[i] = i;
      pi[0] = 0;

      if(spos == RANDOM) curr_node = uniform_int_distribution<>(M, N - 1)(rng);
      if(spos == TAIL)   curr_node = N - 1 - uniform_int_distribution<>(1, log2(N))(rng);
    }
    virtual ~symmetric_app_adversary() {}

    virtual int visit(int v) {
      assert(pi[curr_node] != -1);

      if(not connected(v, curr_node)) return -1;

      if(pi[v] != -1) {
        curr_node = v;
        return ids[pi[v]];
      }

      assert(0 < v and v < M);
      int l = left(curr_node), r = right(curr_node);
      assert(v == l or v == r);
      assert(pi[l] == -1 and pi[r] == -1);
      pi[l] = left(pi[curr_node]); pi[r] = right(pi[curr_node]);
      if(rng() % 2)               swap(pi[l], pi[r]);
      if(in_subtree(pi[v], dest)) swap(pi[l], pi[r]);

      curr_node = v;
      return ids[pi[v]];
    }

  protected:
    int M;
};

class libra_adversary : public basic_adversary {
  public:
    libra_adversary(vector<edge> E, bool left, bool down) : basic_adversary(E) {
      assert(__builtin_popcount(N + 2) == 1);
      M = N >> 1;

      vector<edge> LE, RE;
      for(int i = 0; i < M - 1; ++i) LE.push_back(E[i]);
      assert(edges[M - 1].first == 0 and edges[M - 1].second == M);
      for(int i = M; i < E.size(); ++i) RE.emplace_back(E[i].first - M, E[i].second - M);
      assert(LE.size() == RE.size());

      if(down) {
        assert(left);
        L = new basic_adversary(LE);
        L->curr_node = M - 1;
      }
      else L = new symmetric_adversary(LE, true, false);
      R = new symmetric_adversary(RE, true, false);
      dest = M + R->dest;
      curr_node = (left? L->curr_node : R->curr_node);
    }
    virtual ~libra_adversary() { delete L; delete R; }

    virtual bool answer_bob() { return curr_node >= M and R->answer_bob(); }
    virtual void answer_alice(vector<int> I) {
      basic_adversary::answer_alice(I);

      assert(ids.size() == N);
      vector<int> Lids, Rids;
      for(int i = 0; i < M; ++i) Lids.push_back(ids[i]);
      for(int i = M; i < N; ++i) Rids.push_back(ids[i]);
      assert(Lids.size() == Rids.size());
      L->answer_alice(Lids);
      R->answer_alice(Rids);
    }

    virtual int visit(int v) {
      if(not connected(curr_node, v)) return -1;
      swap(curr_node, v);

      if(curr_node < M) return myvisit(L, curr_node);
      else              return myvisit(R, curr_node - M);
    }

  protected:
    basic_adversary *L, *R;
    int M;

    virtual int myvisit(basic_adversary *a, int v) {
      if(v == a->curr_node) return a->get_id();
      else return a->visit(v);
    }
};

class symmetric_three_adversary : public basic_adversary {
  public:
    symmetric_three_adversary(vector<edge> E) : basic_adversary(E), free_block(2) {
      dest = N / 6;
      curr_node = 0;

      K = 1 + N/3;
      assert(__builtin_popcount(K) == 1);
      assert(edges[0] == make_pair(0, 1));
      assert(edges[1] == make_pair(0, K));
      assert(edges[2] == make_pair(0, K + K - 1));

      int counter = 3;

      for(int t = 0; t < 3; ++t) {
        int off = edges[t].second;

        for(int i = 1; i < K - 1; ++i) {
          assert(edges[counter] == make_pair(i + off, (i - 1) / 2 + off));
          ++counter;
        }
      }

      assert(counter == edges.size());

      pi.resize(N, -1); pi[0] = 0;
    }
    virtual ~symmetric_three_adversary() {}

    virtual bool answer_bob() {
      return pi[curr_node] == dest;
    }

    virtual int visit(int v) {
      if(curr_node == 0) {
        if(v != edges[0].second and v != edges[1].second and v != edges[2].second)
          return -1;

        if(pi[v] == -1) {
          assert(free_block >= 0);
          pi[v] = edges[free_block--].second; // we visit the block actually containing dest last
        }

        curr_node = v;
        return ids[pi[curr_node]];
      }

      if(v == parent(curr_node)) {
        curr_node = v;
        assert(pi[curr_node] != -1);
        return ids[pi[curr_node]];
      }

      if(v != left(curr_node) and v != right(curr_node)) return -1;

      if(pi[v] == -1) {
        int l = left(curr_node), r = right(curr_node);
        assert(pi[l] == -1 and pi[r] == -1);
        assert(pi[curr_node] != -1);
        assert(block(r) == block(curr_node));

        pi[l] = left(pi[curr_node]);
        pi[r] = right(pi[curr_node]);

        if(rng() % 2)               swap(pi[l], pi[r]);
        if(in_subtree(pi[v], dest)) swap(pi[l], pi[r]);
      }

      curr_node = v;
      return ids[pi[curr_node]];
    }

  protected:
    int K;
    int free_block;
    vector<int> pi;

    int block(int x) {
      if(x == 0)        return -1;
      if(x < K)         return  0;
      if(x < K + K - 1) return  1;
      else              return  2;
    }

    int parent(int x) {
      assert(x);
      int off = edges[block(x)].second;
      x -= off;
      if(x == 0) return 0;
                 return (x - 1) / 2 + off;
    }

    int left(int x) {
      assert(x);
      int off = edges[block(x)].second;
      x -= off;
      return 2 * x + 1 + off;
    }

    int right(int x) {
      assert(x);
      int off = edges[block(x)].second;
      x -= off;
      return 2 * x + 2 + off;
    }

    bool in_subtree(int root, int x) {
      while(x > root) x = parent(x);
      return x == root;
    }
};

class line_adversary : public abstract_base_adversary {
  public:
    line_adversary(vector<edge> edges) :
      edges(edges), N(edges.size() + 1), curr(N/2), reversed(false), first(true) {
      assert(N % 2 == 1);
      for(int i = 0; i < edges.size() - 1; ++i)
        assert(edges[i].first == i and edges[i].second == i + 1);
    }
    virtual ~line_adversary() {}

    virtual pair<vector<edge>, int> for_alice() { return {edges, dest}; }
    virtual tuple<vector<edge>, int, int, int> for_bob() { return {edges, curr, ids[curr], N/2}; }

    virtual void answer_alice(vector<int> v) { ids = v; }
    virtual bool answer_bob() { return first? false : curr == get(dest); }
    virtual int visit(int v) {
      if(v != curr + 1 and v != curr - 1) return -1;

      if(first and ids[v] != ids[rev(v)]) {
        first = false;
        reversed = (v < N/2);
      }

      curr = v;

      return ids[get(curr)];
    }

    virtual int get_N() { return N; }

  protected:
    vector<edge> edges;
    int N;
    int curr;
    bool reversed;
    vector<int> ids;
    static const int dest = 0;
    bool first;

    int rev(int x) { return N - 1 - x; }
    int get(int x) { return reversed? rev(x) : x;}
};

template<typename A> class permute_adversary : public abstract_base_adversary {
  public:
    template<typename... Ps> permute_adversary(Ps &&... params) :
      internal_adv(new A(forward<Ps>(params)...)) {}
    virtual ~permute_adversary() { delete internal_adv; }

    virtual pair<vector<edge>, int> for_alice() {
      permute();

      auto [T, dest] = internal_adv->for_alice();
      for(auto &e : T) e = {pi[e.first], pi[e.second]};
      dest = pi[dest];
      shuffle(T.begin(), T.end(), rng);
      return {T, dest};
    }

    virtual tuple<vector<edge>, int, int, int> for_bob() {
      permute();

      auto [T, start, x, dist] = internal_adv->for_bob();
      for(auto &e : T) e = {pi[e.first], pi[e.second]};
      shuffle(T.begin(), T.end(), rng);
      start = pi[start];
      return {T, start, x, dist};
    }

    virtual void answer_alice(vector<int> v) {
      assert(v.size() == get_N());
      vector<int> w(v.size());

      for(int i = 0; i < w.size(); ++i) w[i] = v[pi[i]];
      internal_adv->answer_alice(w);
    }

    virtual bool answer_bob() { return internal_adv->answer_bob(); }
    virtual int visit(int v) { return internal_adv->visit(pi_inv[v]); }
    virtual int get_N() { return internal_adv->get_N(); }

  protected:
    A *internal_adv;
    vector<int> pi;
    vector<int> pi_inv;

    void permute() {
      int N = get_N();

      pi.resize(N);
      iota(pi.begin(), pi.end(), 0);
      shuffle(pi.begin(), pi.end(), rng);

      pi_inv.resize(N);
      for(int i = 0; i < pi.size(); ++i) pi_inv[pi[i]] = i;
    }
};

enum { ALICE = 1337, BOB = 4242 } who;
abstract_base_adversary *adv;
int N;
int Kmax = 0;

void handle_answer_alice() {
  int l = read_or_fail<int>();
  if(l != N) result(0.0, INVALID_ALICE);

  vector<int> ids(N);
  if(fread(ids.data(), sizeof(int), N, fcommin[0]) != N)
    result(0.0, SECURITY_VIOLATION);
  for(int x : ids) {
    if(x < 0) result(0.0, INVALID_ALICE);
    Kmax = max(Kmax, x);
  }

  if(Kmax > _score.back().first) result(0.0, TOO_MANY_TIES);
  adv->answer_alice(ids);
}

int S, counter = 0;

float score(int k) {
  for(auto [bound, s] : _score)
    if(k <= bound) return s;
  assert(false);
  return 0.0;
}

void handle_answer_bob() {
  if(adv->answer_bob()) result(score(Kmax), CORRECT(counter, Kmax));
  else                  result(0.0, WRONG_ANSWER);
}

void send_tree(vector<edge> edges) {
  write(who); write(N);
  for(auto [a, b] : edges) write(min(a, b) + 1), write(max(a, b) + 1);
}

void handle_alice() {
  who = ALICE;

  // Send Alice the tree and the secret node
  auto [T, dest] = adv->for_alice();
  send_tree(T);
  write(dest);
  fflush(fcommout[0]);

  // No need for a loop here: Alice is not allowed to call visit
  int code = read_or_fail<int>();
  if(code == ANSWER)     handle_answer_alice();
  else if(code == VISIT) result(0.0, VISITING_ALICE);
  else                   result(0.0, SECURITY_VIOLATION);
}

void handle_visit() {
  assert(who == BOB);

  ++counter;
  if(counter > S) result(0.0, TOO_MANY_VISITS);

  int w = read_or_fail<int>();
  if(w < 1 or w > N) result(0.0, INVALID_VISIT);
  --w;
  int x = adv->visit(w);
  if(x == -1) result(0.0, INVALID_VISIT);
  else        write(x), fflush(fcommout[0]);
}

void handle_bob() {
  who = BOB;
  restart_submission();

  auto [T, start, x, dist] = adv->for_bob();
  send_tree(T);
  write(start); write(x); fflush(fcommout[0]);
  counter -= dist;

  while(true) {
    int code = read_or_fail<int>();
    if(code == ANSWER)     handle_answer_bob();
    else if(code == VISIT) handle_visit();
    else                   result(0.0, SECURITY_VIOLATION);
  }
}

void check() {
  message_on_shutdown = true;

  tell_cms(1, 1); // no parallel runs, only one test per run
  open_pipes();

  long long seed; int dest;
  assert(fscanf(fin, "%d %d %lld", &N, &S, &seed) == 3);
  rng.seed(seed); rng.discard((35 + N) ^ 4711 ^ 3);

  vector<edge> edges;
  for(int i = 1; i < N; ++i) {
    int u, v; fscanf(fin, "%d %d", &u, &v);
    edges.emplace_back(u, v);
  }

  char buffer[64];
  assert(fscanf(fin, "%s", buffer) == 1);
  string strategy(buffer);

       if(strategy == "basic")       adv = new permute_adversary<basic_adversary>(edges);
  else if(strategy == "sym1")        adv = new permute_adversary<symmetric_adversary>(edges, false, false);
  else if(strategy == "sym2")        adv = new permute_adversary<symmetric_adversary>(edges, true, false);
  else if(strategy == "symo1")       adv = new permute_adversary<symmetric_adversary>(edges, false, true);
  else if(strategy == "symo2")       adv = new permute_adversary<symmetric_adversary>(edges, true, true);
  else if(strategy == "sym3")        adv = new permute_adversary<symmetric_three_adversary>(edges);
  else if(strategy == "sym-app1")    adv = new permute_adversary<symmetric_app_adversary>(edges, true, STD);
  else if(strategy == "sym-app2")    adv = new permute_adversary<symmetric_app_adversary>(edges, false, STD);
  else if(strategy == "sym-app3")    adv = new permute_adversary<symmetric_app_adversary>(edges, true, RANDOM);
  else if(strategy == "sym-app4")    adv = new permute_adversary<symmetric_app_adversary>(edges, true, TAIL);
  else if(strategy == "sym-app5")    adv = new permute_adversary<symmetric_app_adversary>(edges, false, TAIL);
  else if(strategy == "sym-app6")    adv = new permute_adversary<symmetric_app_adversary>(edges, true, STD);
  else if(strategy == "sym-app7")    adv = new permute_adversary<symmetric_app_adversary>(edges, false, STD);
  else if(strategy == "sym-app8")    adv = new permute_adversary<symmetric_app_adversary>(edges, true, RANDOM);
  else if(strategy == "lremap")      adv = new permute_adversary<line_adversary>(edges);
  else if(strategy == "fixed")       adv = new permute_adversary<basic_adversary>(edges, edges.size() / 2, (1 + edges.size()) / 2);
  else if(strategy == "same-center") adv = new permute_adversary<basic_adversary>(edges, 0, 0);
  else if(strategy == "diff-center") adv = new permute_adversary<basic_adversary>(edges, 0, 1);
  else if(strategy == "ftl")         adv = new permute_adversary<basic_adversary>(edges, uniform_int_distribution<>(0, sqrt(edges.size() / 2))(rng), edges.size());
  else if(strategy == "libra1")      adv = new permute_adversary<libra_adversary>(edges, false, false);
  else if(strategy == "libra2")      adv = new permute_adversary<libra_adversary>(edges, true, false);
  else if(strategy == "libra3")      adv = new permute_adversary<libra_adversary>(edges, true, true);
  else                               throw runtime_error("unknown adversary: " + strategy);

  handle_alice();
  handle_bob();
}
